import sys, os
import math
from hanamusume.utils import check_read
        
class OGLMSolver(object):
    def __init__(self, pred=None, bias=True, dim=2**20, alpha=0.1, beta=0.001, l1=0.00001, l2=0.000001):
        self.__name__ = 'OGLMSolver'
        self._dim = dim
        self._w = {}
        self._n = {}
        self._z = {}
        self._a = alpha
        self._b = beta
        self._l1 = l1
        self._l2 = l2
        self._bias = bias
        if pred is None:
            raise AttributeError('OGLMSolver requires a prediction function')
        else:
            self._pred = pred
        
    def predict(self, x):
        dec = 0
        if self._bias:
            x[-1] = 1.
        for i in x:
            if i in self._w:
                dec += self._w[i]*x[i]
        return self._pred(dec)
    
    def update(self, x,p,y):
        if self._bias:
            x[-1] = 1.
        g = (p - y)
        if g == 0:
            return
        for i in x:
            gi = g * x[i]
            self._z[i] = self._z.get(i,0.) + gi
            if i in self._w:
                if i in self._n:
                    ni = self._n[i]
                    self._z[i] -= (1/self._a)*((ni + gi**2)**0.5 - ni**0.5)*self._w[i]
                else:
                    self._z[i] -= gi * self._w[i]
            self._n[i] = self._n.get(i,0.) + gi**2
        for i in x:
            zi = self._z[i]
            ni = self._n[i]
            if abs(zi) > self._l1:
                if zi > 0:
                    self._w[i] = -1/((self._b+ni**0.5)/self._a + self._l2)*(zi-self._l1)
                else:
                    self._w[i] = -1/((self._b+ni**0.5)/self._a + self._l2)*(zi+self._l1)
                    
    def predict_update(self, x,y):
        p = self.predict(x)
        self.update(x, p, y)

    def train_batch(self, data_X, data_y, epoch=20):
        for _ in range(epoch):
           for yy, xx in zip(data_y, data_X):
               self.predict_update(xx,yy)

    def save_model(self, filename, override=False):
        if not override and os.path.exists(filename):
            raise IOError('{} exists'.format(filename))
        i_w = set(self._w.keys())
        i_z = set(self._z.keys())
        i_n = set(self._n.keys())
        assert i_w == i_z
        assert i_w == i_n
        with open(filename,'w') as f:
            f.write('self.__name__:{}\n'.format(self.__name__))
            f.write('self._dim:{}\n'.format(self._dim))
            f.write('self._a:{}\n'.format(self._a))
            f.write('self._b:{}\n'.format(self._b))
            f.write('self._l1:{}\n'.format(self._l1))
            f.write('self._l2:{}\n'.format(self._l2))
            f.write('self._bias:{}\n'.format(self._bias))
            f.write('ind w,z,n:\n')
            for index in sorted(i_w):
                f.write('{} {},{},{}'.format(index, self._w[index], self._z[index], self._n[index]))
                f.write('\n')

    def load_model(self, filename):
        with open(filename) as f:
            check_read(f.readline(),'self.__name__', self.__name__)
            self._dim = int(check_read(f.readline(),'self._dim'))
            self._a = float(check_read(f.readline(),'self._a'))
            self._b = float(check_read(f.readline(),'self._b'))
            self._l1 = float(check_read(f.readline(),'self._l1'))
            self._l2 = float(check_read(f.readline(),'self._l2'))
            self._bias = (check_read(f.readline(),'self._bias') == 'True')
            assert f.readline().strip() == 'ind w,z,n:'
            for line in f.readlines():
                index ,wzn = line.split()
                w,z,n = wzn.split(',')
                index = int(index)
                self._w[index] = float(w)
                self._z[index] = float(z)
                self._n[index] = float(n)



import random

class LibsvmData(object):
    def __init__(self, filename, label_type, label_map=None):
        # for opening file
        self._dataset_filename = filename
        self._dataset_f = None

        # for storing label mapping
        self._ylabel=None
        if label_map == 'auto':
            self.gen_label_map(filename)
        else:
            self.set_label_map(label_map)
        if label_type == 'float':
            self._map_label = lambda x: float(x)
        elif label_type == 'class':
            if self._ylabel:
                self._map_label = lambda x: self._ylabel[x]
            else:
                raise ValueError("label map not set, you can use 'auto' to auto generate from dataset")
        else:
            raise ValueError('unrecognized label type')

        # for generating CV datasets
        self._cv_seed = 1
        self._cv_folds = 5
        self._cv = []

    def __enter__(self):
        self.open_dataset()
        return self

    def open_dataset(self):
        self._dataset_f = open(self._dataset_filename)

    def __exit__(self, type, value, traceback):
        self._dataset_f.close()

    ### Label Map Functions ####
    def set_label_map(self, label_dict):
        self._ylabel = label_dict

    def get_label_map(self):
        return self._ylabel

    def gen_label_map(self, filename):
        with open(filename) as f:
            y_list = [line.split(None,1)[0] for line in f.readlines()]
        labels = sorted(set(y for y in y_list),key=lambda y:float(y))
        if not self._ylabel:
            self._ylabel = {label:float(i) for i,label in enumerate(labels)}
        if any(label not in self._ylabel for label in labels):
            raise ValueError('unrecognized class {}'.format(label))

    #### dataset operation functions ####
    def read_next(self):
        if not self._dataset_f:
            raise ValueError("dataset not opened, use the 'with LibsvmData(<filename>) as data' statement")
        line = self._dataset_f.readline()
        if line:
            y_i, x_i = line.split(None,1)
            y_i = self._map_label(y_i)
            x_i = {int(feature[0]):float(feature[1]) for feature in (comp.split(':') for comp in x_i.split())}
            return (x_i, y_i)
        else:
            return None
    def reset(self):
        self._dataset_f.seek(0)

    #### cv support



import math
from hanamusume.oglm_solver import OGLMSolver

def linear(x):
    return x
class LinearSolver(OGLMSolver):
    def __init__(self, *args, **kwargs):
        kwargs['pred'] = lambda dec: dec
        OGLMSolver.__init__(self, *args, **kwargs)

def sigmoid(x):
    return 1. / (1. + math.exp(-x))
class LogisticSolver(OGLMSolver):
    def __init__(self, *args, **kwargs):
        kwargs['pred'] = lambda dec: sigmoid(dec)
        OGLMSolver.__init__(self, *args, **kwargs)

def exponential(x):
    return math.exp(min(x,5))
class PoissonSolver(OGLMSolver):
    def __init__(self, *args, **kwargs):
        kwargs['pred'] = lambda dec: exponential(x)
        OGLMSolver.__init__(self, *args, **kwargs)



import math
import random
import os

def check_read(line,string,value=None):
    line = line.strip().split(':',1)
    assert len(line) == 2
    assert line[0] == string
    if value is not None:
        assert line[1] == value, '{} {} / {} {}'.format(line[1], type(line[1]), value, type(value))
    return line[1]

def log_loss(probs, labels):
    assert len(probs) == len(labels)
    total_logloss = 0.
    for p,y in zip(probs, labels):
        total_logloss -= y * math.log(p) + (1.-y) * math.log(1.-p)
    return total_logloss/len(probs)

def generate_cv(filename, output_folder=None, n_fold=5, seed=None):
    random.seed(seed)
    if output_folder is None:
        output_folder = '{}.cv'.format(filename)
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    train_f_list = []
    test_f_list = []
    train_fn_list = []
    test_fn_list = []
    for fold in range(n_fold):
        train_fn = '{}/{}.{}.tr'.format(output_folder, filename, fold) 
        test_fn = '{}/{}.{}.ts'.format(output_folder, filename, fold)
        train_fn_list.append(train_fn)
        test_fn_list.append(test_fn)
        train_f_list.append(open(train_fn,'w'))
        test_f_list.append(open(test_fn,'w'))
    
    with open(filename) as f:
        for line in f.readlines():
            ind = random.randint(0,n_fold-1)
            for i in range(n_fold):
                if i == ind:
                    test_f_list[i].write(line)
                else:
                    train_f_list[i].write(line)
    for f in train_f_list:
        f.close()
    for f in test_f_list:
        f.close()

    return train_fn_list, test_fn_list
